'This code launches a new cog to generate high quality sine waves at one or two frequencies with variable duration and volume.
'Copyright 2009 Raymond Allen
'
VAR 'Tone parameters  (do not change the order of these variables and keep them together)
  long    ToneNumberSteps
  long    ToneLeftPhaseStep
  long    ToneRightPhaseStep
  long    ToneLeftAttenuation
  long    ToneRightAttenuation
  long    ToneLeftPin
  long    ToneRightPin
  long    ToneRunning     'this flag is set to zero when assembly routine finished and cog stopped

PUB ToneGen(frequency, duration, attenuation) |LeftFrequency,RightFrequency
  'Set pin numbers for output (PSM uses pins 4&5, demo board uses pins 10 & 11)
  'Set either to a negative number if you don't want to use it
  ToneLeftPin:=5
  ToneRightPin:=4 

  'Set desired frequency (in Hz) for left & right channels
  LeftFrequency:=frequency'440
  RightFrequency:=frequency'445

  'Set desired time to ouput tone
  'duration:=2000  'in milliseconds  (1000=1 second)

  'Convert tone duration into the total number of main loop cycles for the assembly routine
  ToneNumberSteps:=(duration*(clkfreq/1000))/200')  = time/(time/step) = (ms/1000)/(100/clkfreq)
  'Number of Steps= ((1000*seconds)*(clocks/second)/1000)/(steps/cycle)
  'We calculate the total number of main loop cycles needed to give the desired output time here
  '(because the math is easier in SPIN than in assembly!)
  '(1000*seconds) is the requested tone duration in milliseconds
  'clocks/second = clkfreq (the Propeller clock frequency is normally 80,000,000 Hz)
  '1000 is divided to turn milliseconds into seconds
  'clocks/step = 200  (output is updated every 200 clock cycles (assembly main loop time), this is a 400 kHz update rate)
  

  'Convert frequencies into a phase step for the assembly routine
  'ToneLeftPhaseStep:=  (LeftFrequency*(1<<19)/(clkfreq/200))*(1<<13) ' '8192=2^13=input range of sin function
  'ToneRightPhaseStep:=(RightFrequency*(1<<19)/(clkfreq/200))*(1<<13) ' '524288=2^19
  ToneLeftPhaseStep:=  (LeftFrequency*(1<<15)/(clkfreq/200))*(1<<13)*(1<<4) ' '8192=2^13=input range of sin function
  ToneRightPhaseStep:=(RightFrequency*(1<<15)/(clkfreq/200))*(1<<13)*(1<<4) ' '524288=2^19
  'Phase step= (cycles/second)*(clocks/step)*(steps/cycle)/(clocks/second)
  'We convert the input frequency into a phase step here because the math is easier in SPIN than in Assembly!
  '(In the assembly main loop, the phase accumulator registers are increased by this value every loop)
  'cycles/second = Hz = desired frequency
  'clocks/step = 200  (output is updated every 200 clock cycles (assembly main loop time), this is a 400 kHz update rate)
  'steps/cycle = 8192 = 1<<13 (this input range to the assembly sin subroutine represents one full cycle of a sine wave)
  'clocks/second = clkfreq (the Propeller clock frequency is normally 80,000,000 Hz)
  'Note: Phase step is shifted left (1<<19) here and then phase is shifted right the same amount in the assembly section
  '  (this adds a bit to the precision and simplifies the phase calculation)

  'Select Attenuations (0=loud, 8=soft) 
  ToneLeftAttenuation:=attenuation
  ToneRightAttenuation:=attenuation
  '(The final values are shifted right this many times)


  ToneRunning:=true    'Assembly routine will set this false when done 
  cognew(@ToneStart,@ToneNumberSteps)  'Start a cog to generate the tones




DAT  'Assembly Tone Generator
org 0
ToneStart  'Beginning of assembly tone generator
'load input parameters from hub to cog given address in par
              movd      :par,#tNumberSteps        'copies 7 parameters to location starting with tNumberSteps  
              mov       temp1,par
              mov       temp2,#7  'there are input 7 parameters
:par          rdlong    0,temp1
              add       :par,dlsb
              add       temp1,#4
              djnz      temp2,#:par

              mov       temp2,temp1  'save next address because it's the output address to say when we're done



'setup output pins (make them outputs unless they are negative numbers)              
RightPinSetup
              TEST      tRightPin,offset  wz 'See if right pin # is a negative number (if so, do not enable output)
        if_z  MOV       temp1,#1
        if_z  ROL       temp1,tRightPin
        if_z  OR        DIRA, temp1
LeftPinSetup
              TEST      tLeftPin,offset   wz  'See if left pin # is a negative number (if so, do not enable output)      
        if_z  MOV       temp1,#1
        if_z  ROL       temp1,tLeftPin
        if_z  OR        DIRA, temp1

'setup counters (A for right pin, B for left pin)
              mov       temp1,tCounterMode
              OR        temp1,tRightPin
              MOV       CTRA,temp1
              mov       temp1,tCounterMode
              OR        temp1,tLeftPin
              MOV       CTRB,temp1


'Setup for main loop               
              mov       wait,cnt
              add       wait,dcnt

'Main Loop
ToneLoop
              waitcnt   wait,dcnt   'this waitcnt ensures that every loop is exactly 200 clock cycles

              'Update left channel
              'calculate new phase
              add       LeftPhase,tLeftStep
              mov       sin,LeftPhase
              shr       sin,#19    'shifts back into position for sin subroutine
              'get sine value from phase
              call      #GetSin   'call the getsin subroutine
              'shift into position and apply attenuation
              shl       sin,#15
              SAR       sin,tLeftAtten
              add       sin,offset   'this puts zero at midlevel output
              mov       frqa,sin

              'Update right channel
              'calculate new phase 
              add       RightPhase,tRightStep
              mov       sin,RightPhase
              shr       sin,#19   'shifts back into position for sin subroutine
              'get sine value from phase 
              call      #GetSin    'call the getsin subroutine  
              shl       sin,#15
              SAR       sin,tRightAtten
              add       sin,offset   'this puts zero at midlevel output            
              mov       frqb,sin
              
              'Decrement #step counter and repeat main loop until step counter is down to zero
              DJNZ      tNumberSteps,#ToneLoop

'End of main loop
              
'Stop cog (because we're all done)
              mov       temp1,#0   'temp1=0 (false)
              wrlong    temp1,temp2    'Clear flag to indicate still running              
              COGID     temp1
              COGSTOP   temp1              

DAT  'This sine routine is straight out of the manual.
              ' Get sine/cosine
              '
              ' quadrant: 1 2 3 4
              ' angle: $0000..$07FF $0800..$0FFF $1000..$17FF $1800..$1FFF
              ' table index: $0000..$07FF $0800..$0001 $0000..$07FF $0800..$0001
              ' mirror: +offset -offset +offset -offset
              ' flip: +sample +sample -sample -sample
              '
              ' on entry: sin[12..0] holds angle (0 to just under 360)
              ' on exit: sin holds signed value ranging from $0000FFFF ('1') to
              ' $FFFF0001 ('-1')
getcos        add       sin,sin_90 'for cosine, add 90
getsin        test      sin,sin_90 wc 'get quadrant 2|4 into c
              test      sin,sin_180 wz 'get quadrant 3|4 into nz
              negc      sin,sin 'if quadrant 2|4, negate offset
              or        sin,sin_table 'or in sin table address >> 1
              shl       sin,#1 'shift left to get final word address
              rdword    sin,sin 'read word sample from $E000 to $F000
              negnz     sin,sin 'if quadrant 3|4, negate sample
getsin_ret
getcos_ret    RET '39..54 clocks
                        '(variance due to HUB sync on RDWORD)
DAT  'variables for tone generation
'Data for Sine Table Extraction
sin_90        long $0800
sin_180       long $1000
sin_table     long $E000 >> 1 'sine table base shifted right
sin           long 0
sinmask       long $0000_1FFF



'setup parameters                           
tCounterMode  long %00011000_00000000_00000000_00000000
dlsb          long  1 << 9

'phase accumulators
LeftPhase     long  0
RightPhase    long  0

'main loop vars
dcnt          long  200  '#clocks/main loop  --> gives 400kHz update rate
offset        long  $8000_0000

'Working variables
temp1         res  1
temp2         res  1          
wait          res  1

'Input Parameters
tNumberSteps  res  1    'Determines tone duration (left and right)
tLeftStep     res  1    'Determines left tone frequency 
tRightStep    res  1    'Determines right tone frequency    
tLeftAtten    res  1    'Determines left tone volume (0=loud, 8=soft)
tRightAtten   res  1    'Determines right tone volume (0=loud, 8=soft) 
tLeftPin      res  1    'Pin# for left audio output (negative number disables)
tRightPin     res  1    'Pin# for right audio output (negative number disables)